import CADETMatch.util as util
import CADETMatch.score as score
import scipy.stats
import scipy.interpolate
import numpy
from addict import Dict
import CADETMatch.smoothing as smoothing

name = "ShapeDecay"
settings = Dict()
settings.adaptive = True
settings.badScore = 0
settings.meta_mask = True
settings.count = 6
settings.failure = [0.0] * settings.count, 1e6, 1, numpy.array([0.0]), numpy.array([0.0]), numpy.array([1e6]), [1.0] * settings.count

def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)
    selected = feature['selected']

    exp_data_values = feature['value'][selected]
    exp_time_values = feature['time'][selected]

    sim_data_values_spline = smoothing.smooth_data_derivative(exp_time_values, sim_data_values, feature['critical_frequency'], feature['smoothing_factor'])
    exp_data_values_spline = smoothing.smooth_data_derivative(exp_time_values, exp_data_values, feature['critical_frequency'], feature['smoothing_factor'])
     
    [high, low] = util.find_peak(exp_time_values, sim_data_values)

    time_high, value_high = high

    pearson, diff_time = score.pearson_spline(exp_time_values, sim_data_values, exp_data_values)

    pearson_der, diff_time_der = score.pearson_spline(exp_time_values, sim_data_values_spline, exp_data_values_spline)

    [highs_der, lows_der] = util.find_peak(exp_time_values, sim_data_values_spline)

    
    temp = [pearson, 
            feature['value_function'](value_high), 
            feature['time_function'](numpy.abs(diff_time)),
            pearson_der,
            feature['value_function_high'](highs_der[1]),             
            feature['value_function_low'](lows_der[1]),]
    return (temp, util.sse(sim_data_values, exp_data_values), len(sim_data_values), 
            sim_time_values, sim_data_values, exp_data_values, [1.0 - i for i in temp])

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    name = '%s_%s' % (sim.root.experiment_name,   feature['name'])
    s, crit_fs = smoothing.find_smoothing_factors(selectedTimes, selectedValues, name, cache)
    values = smoothing.smooth_data_derivative(selectedTimes, selectedValues, crit_fs, s)

    [high, low] = util.find_peak(selectedTimes, values)

    temp = {}
    temp['peak'] = util.find_peak(selectedTimes, selectedValues)[0]
    
    temp['time_function'] = score.time_function_decay_cv(CV_time, selectedTimes, temp['peak'][0])
    temp['value_function'] = score.value_function(temp['peak'][1], abstol)
    temp['value_function_high'] = score.value_function(high[1], numpy.abs(high[1])/1000)
    temp['value_function_low'] = score.value_function(low[1], numpy.abs(low[1])/1000)
    temp['peak_max'] = max(selectedValues)
    temp['smoothing_factor'] = s
    temp['critical_frequency'] = crit_fs
    return temp

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_Similarity" % name, "%s_Value" % name, "%s_Time" % name,
            "%s_Derivative_Similarity" % name,
            "%s_Der_High_Value" % name, "%s_Der_Low_Value" % name]
    return temp





import importlib
import importlib.util
import sys

from pathlib import Path

base = Path(__file__).parent

def load_plugin(path):
    module = '.'.join(path.relative_to(base).parts).replace('.py', '')
    spec = importlib.util.spec_from_file_location(module, str(path))
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    sys.modules[foo.__name__] = foo
    return foo

def get_files(dir):
    return [path for path in dir.glob('*.py') if not path.name == '__init__.py']

def get_plugins(directory):
    plugins = base / directory
    temp = {}
    for path in get_files(plugins):
        plug = load_plugin(path)
        temp[plug.name] = plug
    return temp



import numpy

def linear_coeff(x1, y1, x2, y2):
    "return paramters that fit y = a*x+b"
    a = (y2 - y1)/(x2 - x1)
    b = y2 - a * x2
    return a,b

def exponential_coeff(x1, y1, x2, y2):
    "return paramaters that fit y = b*exp(m*x)"
    b = (numpy.log(y2) - numpy.log(y1))/(x2 - x1)
    a = y1 * numpy.exp(-b*x1)
    return a,b

def exponential(x, a, b):
    return a * numpy.exp(b*x)

def linear(x, a, b):
    return a*x+b




import CADETMatch.util as util
import CADETMatch.score as score
import numpy
from addict import Dict

name = "width"
settings = Dict()
settings.adaptive = True
settings.badScore = 0
settings.meta_mask = True
settings.count = 3
settings.failure = [0.0] * settings.count, 1e6, 1, numpy.array([0.0]), numpy.array([0.0]), numpy.array([1e6]), [1.0] * settings.count

def run(sim_data, feature):
    "similarity, value, start stop"
    sim_time_values, sim_data_values = util.get_times_values(sim_data['simulation'], feature)
    selected = feature['selected']

    exp_data_values = feature['value'][selected]
    exp_time_values = feature['time'][selected]

    sim_width_25 = find_width(sim_time_values, sim_data_values, 0.25)
    sim_width_50 = find_width(sim_time_values, sim_data_values, 0.50)
    sim_width_75 = find_width(sim_time_values, sim_data_values, 0.75)
    
    temp = [feature['width_25'](sim_width_25), 
            feature['width_50'](sim_width_50), 
            feature['width_75'](sim_width_75),]
    return (temp, util.sse(sim_data_values, exp_data_values), len(sim_data_values), 
            sim_time_values, sim_data_values, exp_data_values, [1.0 - i for i in temp])

def setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, cache):
    temp = {}
    temp['width_25'] = score.value_function(find_width(selectedTimes, selectedValues, 0.25), abstol)
    temp['width_50'] = score.value_function(find_width(selectedTimes, selectedValues, 0.50), abstol)
    temp['width_75'] = score.value_function(find_width(selectedTimes, selectedValues, 0.75), abstol)
    temp['peak_max'] = max(selectedValues)
    return temp

def find_width(times, values, percent):
    idx_max = numpy.argmax(values)

    max_value = values[idx_max]
    max_time = times[idx_max]
    
    idx_upper = numpy.argmax(values[idx_max:] <= (max_value * percent))
    
    idx_lower = numpy.argmin(values[:idx_max] <= (max_value * percent))
    
    diff_time = times[idx_max + idx_upper] - times[idx_lower]
    
    return diff_time

def headers(experimentName, feature):
    name = "%s_%s" % (experimentName, feature['name'])
    temp = ["%s_width_25" % name, "%s_width_50" % name, "%s_width_75" % name]
    return temp





import CADETMatch.util as util
import math

name = "norm_diameter"
count = 1
count_extended = 1

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    minValue = parameter['min']
    maxValue = parameter['max']

    def trans(i):
        return (i - minValue)/(maxValue-minValue)

    return [trans,]

def untransform(seq, cache, parameter):
    minValue = parameter['min']
    maxValue = parameter['max']

    values_transform = [(maxValue - minValue) * seq[0] + minValue,]

    values = [math.pi * values_transform[0]**2/4.0,]
    headerValues = [values[0], values_transform[0]]
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    values = untransform_matrix_inputorder(matrix, cache, parameter)
    values[:,0] = math.pi * values[:,0]**2/4.0
    return values

def untransform_matrix_inputorder(matrix, cache, parameter):
    minValue = parameter['min']
    maxValue = parameter['max']

    values = (maxValue - minValue) * matrix + minValue
    return values

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)

    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        location = parameter['location']
    
        try:
            comp = parameter['component']
            bound = parameter['bound']
            index = None
        except KeyError:
            index = parameter['index']
            bound = None

        if bound is not None:
            unit = getUnit(location)
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            if comp == -1:
                position = ()
                sim[location.lower()] = values[0]
            else:
                position = boundOffset[comp] + bound
                sim[location.lower()][position] = values[0]

        if index is not None:
            sim[location.lower()][index] = values[0]
    return values, headerValues

def setupTarget(parameter):
    location = parameter['location']
    bound = parameter['bound']
    comp = parameter['component']

    name = location.rsplit('/', 1)[-1]
    sensitivityOk = 1

    try:
        unit = int(location.split('/')[3].replace('unit_', ''))
    except ValueError:
        unit = ''
        sensitivityOk = 0

    return [(name, unit, comp, bound),], sensitivityOk

def getBounds(parameter):
    return [0.0,], [1.0,]

def getHeaders(parameter):
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("Area Comp:%s Bound:%s" % (comp, bound))
    headers.append("Diameter Comp:%s Bound:%s" % (comp, bound))
    return headers

def getHeadersActual(parameter):
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("Area Comp:%s Bound:%s" % (comp, bound))
    return headers

def setBounds(parameter, lb, ub):
    parameter['min'] = lb[0]
    parameter['max'] = ub[0]




from deap import tools
from operator import eq

class ParetoFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None, similar_fit = None):
        if similar is None:
            similar = eq
        if similar_fit is not None:
            self.similar_fit = similar_fit
        else:
            self.similar_fit = eq
        super().__init__(similar)

    def update(self, population):
        """Update the Pareto front hall of fame with the *population* by adding 
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.
        
        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []
            for i, hofer in enumerate(self):    # hofer = hall of famer
                if not dominates_one and hofer.fitness.dominates(ind.fitness):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hofer.fitness):
                    dominates_one = True
                    to_remove.append(i)
                elif self.similar_fit(ind.fitness.values, hofer.fitness.values) and self.similar(ind, hofer):
                    has_twin = True
                    break
            
            for i in reversed(to_remove):       # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                self.insert(ind)

class DummyFront(tools.ParetoFront):
    "Modification of the pareto front in DEAP that takes cache as an argument to update to use for similar comparison"

    def __init__(self, similar=None):
        "This is here for API compatibility, don't do anything"
        if similar is None:
            similar = eq
        super().__init__(similar)

    def update(self, population):
        "do not put anything in this front, it is just needed to maintain compatibility"
        pass


import numpy
from sklearn import preprocessing
from sklearn.neighbors.kde import KernelDensity
from sklearn.model_selection import cross_val_score
import scipy
import multiprocessing
import sys

import cadet
import CADETMatch.util as util
import CADETMatch.evo as evo
import pandas
from addict import Dict
from CADETMatch.cache import cache
from pathlib import Path
import warnings
import joblib
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py

import matplotlib
matplotlib.use('Agg')

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

import matplotlib.pyplot as plt

import matplotlib.cm
cm_plot = matplotlib.cm.gist_rainbow

import logging
import CADETMatch.loggerwriter as loggerwriter

import CADETMatch.kde_generator as kde_generator

def get_color(idx, max_colors, cmap):
    return cmap(1.*float(idx)/max_colors)

saltIsotherms = {b'STERIC_MASS_ACTION', b'SELF_ASSOCIATION', b'MULTISTATE_STERIC_MASS_ACTION', 
                 b'SIMPLE_MULTISTATE_STERIC_MASS_ACTION', b'BI_STERIC_MASS_ACTION'}

size = 20

plt.rc('font', size=size)          # controls default text sizes
plt.rc('axes', titlesize=size)     # fontsize of the axes title
plt.rc('axes', labelsize=size)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=size)    # fontsize of the tick labels
plt.rc('ytick', labelsize=size)    # fontsize of the tick labels
plt.rc('legend', fontsize=size)    # legend fontsize
plt.rc('figure', titlesize=size)  # fontsize of the figure title
plt.rc('figure', autolayout=True)

def plotTube(cache, chain, kde, scaler):
    results, combinations = processChainForPlots(cache, chain, kde, scaler)

    output_mcmc = cache.settings['resultsDirSpace'] / "mcmc"
    output_mcmc.mkdir(parents=True, exist_ok=True)

    mcmc_h5 = output_mcmc / "mcmc_plots.h5"

    h5 = cadet.H5()
    h5.filename = mcmc_h5.as_posix()

    for expName,value in combinations.items():
        exp_name = expName.split('_')[0]
        plot_mcmc(output_mcmc, value, expName, "combine", cache.target[exp_name]['time'], cache.target[exp_name]['value'])
        h5.root[expName] = value['data']
        h5.root['exp_%s_time' % expName] = cache.target[exp_name]['time']
        h5.root['exp_%s_value' % expName] = cache.target[exp_name]['value']

    for exp, units in results.items():
        for unitName, unit in units.items():
            for comp, data in unit.items():
                expName = '%s_%s' % (exp, unitName)
                plot_mcmc(output_mcmc, data, expName, comp, cache.target[exp]['time'], cache.target[exp]['value'])
                h5.root['%s_%s' % (expName, comp)] = data['data']
                h5.root['exp_%s_%s_time' % (expName, comp)] = cache.target[exp_name]['time']
                h5.root['exp_%s_%s_value' % (expName, comp)] = cache.target[exp_name]['value']
    h5.save()

def processChainForPlots(cache, chain, kde, scaler):
    mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, results, times, mcmc_score = genRandomChoice(cache, chain, kde, scaler)

    writeSelected(cache, mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, mcmc_score)

    results, combinations = processResultsForPlotting(results, times)

    return results, combinations

def processResultsForPlotting(results, times):
    for expName, units in list(results.items()):
        for unitName, unit in list(units.items()):
            for compIdx, compValue in list(unit.items()):
                data = numpy.array(compValue)
                results[expName][unitName][compIdx] = {}
                results[expName][unitName][compIdx]['data'] = data
                results[expName][unitName][compIdx]["mean"] = numpy.mean(data, 0)
                results[expName][unitName][compIdx]["std"] = numpy.std(data, 0)
                results[expName][unitName][compIdx]["min"] = numpy.min(data, 0)
                results[expName][unitName][compIdx]["max"] = numpy.max(data, 0)
                results[expName][unitName][compIdx]['time'] = times[expName]

    combinations = {}
    for expName, units in results.items():
        for unitName, unit in list(units.items()):
            data = numpy.zeros(unit[0]['data'].shape)
            times = unit[0]['time']
            for compIdx, compValue in list(unit.items()):
                data = data + compValue['data']
            temp = {}
            temp['data'] = data
            temp['time'] = times
            temp["mean"] = numpy.mean(data, 0)
            temp["std"] = numpy.std(data, 0)
            temp["min"] = numpy.min(data, 0)
            temp["max"] = numpy.max(data, 0)
            comb_name = '%s_%s' % (expName, unitName)
            combinations[comb_name] = temp
    return results, combinations

def fitness(individual):
    return evo.fitness(individual, sys.argv[1])

def genRandomChoice(cache, chain, kde, scaler):
    "want about 1000 items and will be removing about 10% of them"
    size = 1100
    chain = chain[~numpy.all(chain == 0, axis=1)]
    if len(chain) > size:
        indexes = numpy.random.choice(chain.shape[0], size, replace=False)
        chain = chain[indexes,:]

    lb, ub = numpy.percentile(chain, [5, 95], 0)
    selected = (chain >= lb) & (chain <= ub)
    bools = numpy.all(selected, 1)
    chain = chain[bools, :]

    individuals = []

    for idx in range(len(chain)):
        individuals.append(chain[idx,:])

    map_function = util.getMapFunction()
    fitnesses = map_function(fitness, individuals)

    results = {}
    times = {}

    mcmc_selected = []
    mcmc_selected_transformed = []
    mcmc_selected_score = []

    for (fit, csv_line, result) in fitnesses:        
        if result is not None:
            mcmc_selected_score.append(tuple(fit))
            for value in result.values():
                mcmc_selected_transformed.append(tuple(value['cadetValues']))
                mcmc_selected.append(tuple(value['individual']))
                break
  
            for key,value in result.items():
                sims = results.get(key, {})

                sim = value['simulation']

                outlets = [(name, sim.root.input.model[name].ncomp) for name in cache.target[key]['units_used']]

                for outlet, ncomp in outlets:
                    units = sims.get(outlet, {})

                    for i in range(ncomp):
                        comps = units.get(i, [])
                        comps.append(sim.root.output.solution[outlet]["solution_outlet_comp_%03d" % i])
                        units[i] = comps

                    sims[outlet] = units

                    if key not in times:
                        times[key] = sim.root.output.solution.solution_times

                results[key] = sims
        else:
            multiprocessing.get_logger().info("Failure in random choice: fit: %s  csv_line: %s   result:%s", fit, csv_line, result)

    mcmc_selected = numpy.array(mcmc_selected)
    mcmc_selected_transformed = numpy.array(mcmc_selected_transformed)
    mcmc_selected_score = numpy.array(mcmc_selected_score)

    mcmc_score = kde.score_samples(scaler.transform(mcmc_selected_score)) + numpy.log(2)

    return mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, results, times, mcmc_score

def writeSelected(cache, mcmc_selected, mcmc_selected_transformed, mcmc_selected_score, mcmc_score):
    mcmc_h5 = Path(cache.settings['resultsDirMCMC']) / "mcmc_selected.h5"
    h5 = cadet.H5()
    h5.filename = mcmc_h5.as_posix()
    h5.root.mcmc_selected = numpy.array(mcmc_selected)
    h5.root.mcmc_selected_transformed = numpy.array(mcmc_selected_transformed)
    h5.root.mcmc_selected_score = numpy.array(mcmc_selected_score)
    h5.root.mcmc_selected_kdescore = numpy.array(mcmc_score)
    h5.save()

def plot_mcmc(output_mcmc, value, expName, name, expTime, expValue):
    data = value['data']
    times = value['time']
    mean = value["mean"]
    std = value["std"]
    minValues = value["min"]
    maxValues = value["max"]

    plt.figure(figsize=[10,10])
    plt.plot(times, mean, label='mean')
    plt.fill_between(times, mean - std, mean + std,
                color='green', alpha=0.2, label='high prob')
    plt.fill_between(times, minValues, maxValues,
                color='red', alpha=0.2, label='low prob')
    plt.plot(expTime, expValue, 'r', label='exp')
    plt.xlabel('time(s)')
    plt.ylabel('conc(mM)')
    plt.legend()
    plt.savefig(str(output_mcmc / ("%s_%s.png" % (expName, name) ) ))
    plt.close()

    row, col = data.shape
    alpha = 0.01
    plt.figure(figsize=[10,10])
    plt.plot(times, data[0,:], 'g', alpha=0.5, label='prob')
    plt.plot(times, data.transpose(), 'g', alpha=alpha)
    plt.plot(times, mean, 'k', label='mean')
    plt.plot(expTime, expValue, 'r', label='exp')
    plt.xlabel('time(s)')
    plt.ylabel('conc(mM)')
    plt.legend()

    plt.savefig(str(output_mcmc / ("%s_%s_lines.png" % (expName, name) ) ))
    plt.close()

def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings['resultsDirLog'], "mcmc_plot_tube.log")
    cache.setup(sys.argv[1])

    kde, kde_scaler = kde_generator.getKDE(cache)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    with h5py.File(mcmc_h5, 'r') as h5:
        flat_chain = h5['/flat_chain'][()]
        plotTube(cache, flat_chain, kde, kde_scaler)

if __name__ == "__main__":
    main()




import numpy
from sklearn import preprocessing
from sklearn.neighbors.kde import KernelDensity
from sklearn.model_selection import cross_val_score
import scipy
import multiprocessing
import sys

import cadet
import CADETMatch.util as util
import CADETMatch.evo as evo
import pandas
from addict import Dict
from CADETMatch.cache import cache
import CADETMatch.smoothing as smoothing
from pathlib import Path
import warnings
import joblib
with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=FutureWarning)
    import h5py

import matplotlib
matplotlib.use('Agg')

from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

import matplotlib.pyplot as plt

import matplotlib.cm
cm_plot = matplotlib.cm.gist_rainbow

import logging
import CADETMatch.loggerwriter as loggerwriter

def get_color(idx, max_colors, cmap):
    return cmap(1.*float(idx)/max_colors)

saltIsotherms = {b'STERIC_MASS_ACTION', b'SELF_ASSOCIATION', b'MULTISTATE_STERIC_MASS_ACTION', 
                 b'SIMPLE_MULTISTATE_STERIC_MASS_ACTION', b'BI_STERIC_MASS_ACTION'}

size = 20

plt.rc('font', size=size)          # controls default text sizes
plt.rc('axes', titlesize=size)     # fontsize of the axes title
plt.rc('axes', labelsize=size)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=size)    # fontsize of the tick labels
plt.rc('ytick', labelsize=size)    # fontsize of the tick labels
plt.rc('legend', fontsize=size)    # legend fontsize
plt.rc('figure', titlesize=size)  # fontsize of the figure title
plt.rc('figure', autolayout=True)

def reduce_data(data, size):
    lb, ub = numpy.percentile(data, [5, 95], 0)
    selected = (data >= lb) & (data <= ub)

    selected = numpy.all(selected,1)

    data = data[selected,:]

    shape = data.shape

    if size < data.shape[0]:
        indexes = numpy.random.choice(data.shape[0], size, replace=False)
        data_reduced = data[indexes]
    else:
        data_reduced = data

    scaler = preprocessing.StandardScaler().fit(data)
    data = scaler.transform(data)

    data_reduced = scaler.transform(data_reduced)

    return data, data_reduced, scaler

def bandwidth_score(bw, data, kernel, atol):
    bandwidth = 10**bw[0]
    kde_bw = KernelDensity(kernel=kernel, bandwidth=bandwidth, atol=atol)
    scores = cross_val_score(kde_bw, data, cv=3)
    return -max(scores)

def goal_kde(x, kde):
    test_value = numpy.array(x).reshape(1, -1)
    score = kde.score_samples(test_value)
    return -score[0]

def get_mle(data):
    atol = 1e-4
    kernel = 'gaussian'

    multiprocessing.get_logger().info("setting up scaler and reducing data")
    data, data_reduced, scaler = reduce_data(data, 32000)
    multiprocessing.get_logger().info("finished setting up scaler and reducing data")
    multiprocessing.get_logger().info('data_reduced shape %s', data_reduced.shape)

    BOUND_LOW_num = numpy.min(data, 0)
    BOUND_UP_num = numpy.max(data, 0)

    BOUND_LOW_trans = list(BOUND_LOW_num)
    BOUND_UP_trans = list(BOUND_UP_num)    

    BOUND_LOW_real = [0.0] * len(BOUND_LOW_trans)
    BOUND_UP_real = [1.0] * len(BOUND_UP_trans)

    map_function = util.getMapFunction()

    result = scipy.optimize.differential_evolution(bandwidth_score, bounds = [(-3, 1),], 
                                               args = (data_reduced, kernel, atol), 
                                               updating='deferred', workers=map_function, disp=True,
                                               popsize=50)
    bw = 10**result.x[0]
    
    multiprocessing.get_logger().info("mle bandwidth: %.2g", bw)

    kde_ga = KernelDensity(kernel=kernel, bandwidth=bw, atol=atol)

    multiprocessing.get_logger().info('fitting kde with mle bandwidth')
    kde_ga = kde_ga.fit(data_reduced)
    multiprocessing.get_logger().info('finished fitting and starting mle search')
        
    result_kde = scipy.optimize.differential_evolution(goal_kde, bounds = list(zip(BOUND_LOW_trans, BOUND_UP_trans)), 
                                               args = (kde_ga,), 
                                               updating='deferred', workers=map_function, disp=True,
                                               popsize=50)
    multiprocessing.get_logger().info('finished mle search')

    x = list(scaler.inverse_transform(numpy.array(result_kde.x).reshape(1, -1))[0])

    multiprocessing.get_logger().info('mle found %s', x)

    return x, kde_ga, scaler

def fitness(individual):
    return evo.fitness(individual, sys.argv[1])

def addChain(axis, *args):
    temp = [arg for arg in args if len(arg)]
    if len(temp) > 1:
        return numpy.concatenate( temp, axis=axis)
    else:
        return numpy.array(temp[0])

def process_mle(chain, gen, cache):
    mcmc_dir = Path(cache.settings['resultsDirMCMC'])

    mcmc_csv = Path(cache.settings['resultsDirMCMC']) / "prob.csv"
    mle_h5 = Path(cache.settings['resultsDirMCMC']) / "mle.h5"

    h5 = cadet.H5()
    h5.filename = mle_h5.as_posix()
    if mle_h5.exists():
        h5.load()

        if h5.root.generations[-1] == gen:
            multiprocessing.get_logger().info('new information is not yet available and mle will quit')
            return

    multiprocessing.get_logger().info('process mle chain shape before %s', chain.shape)
    #This step cleans up bad entries
    chain = chain[~numpy.all(chain == 0, axis=1)]
    multiprocessing.get_logger().info('process mle chain shape after cleaning 0 entries %s', chain.shape)
    mle_x, kde, scaler = get_mle(chain)
    multiprocessing.get_logger().info("mle_x: %s", mle_x)

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    joblib.dump(scaler, mcmcDir / 'kde_prior_scaler.joblib')
    joblib.dump(kde, mcmcDir / 'kde_prior.joblib')

    mle_ind = util.convert_individual(mle_x, cache)[0]
    
    multiprocessing.get_logger().info("mle_ind: %s", mle_ind)

    temp = [mle_x,]

    multiprocessing.get_logger().info('chain shape: %s', chain.shape)

    #run simulations for 5% 50% 95% and MLE vs experimental data
    percentile_splits = [5, 10, 50, 90, 95]
    percentile = numpy.percentile(chain, percentile_splits, 0)

    multiprocessing.get_logger().info("percentile: %s %s", percentile.shape, percentile)

    for row in percentile:
        temp.append(list(row))

    cadetValues = [util.convert_individual(i, cache)[0] for i in temp]
    cadetValues = numpy.array(cadetValues)

    multiprocessing.get_logger().info('cadetValues: %s %s', cadetValues.shape, cadetValues)

    map_function = util.getMapFunction()

    fitnesses = list(map_function(fitness, temp))

    simulations = {}
    for scores, csv_record, results in fitnesses:
        for name, value in results.items():
            sims = simulations.get(name, [])
            sims.append(value['simulation'])

            simulations[name] = sims
                       
    multiprocessing.get_logger().info('type %s  value %s', type(cadetValues), cadetValues)

    pd = pandas.DataFrame(cadetValues, columns = cache.parameter_headers_actual)
    labels = ['MLE', '5', '10', '50', '90', '95']
    pd.insert(0, 'name', labels)
    pd.to_csv(mcmc_csv, index=False)

    h5.root.stat_labels = cache.parameter_headers_actual
    h5.root.percentile_splits = percentile_splits

    mle_x = numpy.array(mle_x)

    h5.root.mles = addChain(1, h5.root.mles, mle_x[:,numpy.newaxis])
    h5.root.stats = addChain(2, h5.root.stats, percentile[:, :, numpy.newaxis])
    h5.root.generations = addChain(0, h5.root.generations, [gen,])
    h5.root.stat_MLE = mle_ind

    h5.save()

    plot_mle(simulations, cache, labels)

def plot_mle(simulations, cache, labels):
    mcmc_dir = Path(cache.settings['resultsDirMCMC'])
    target = cache.target
    settings = cache.settings
    for experiment in settings['experiments']:
        experimentName = experiment['name']
        
        file_name = '%s_stats.png' % experimentName
        dst = mcmc_dir / file_name

        numPlotsSeq = [1]
        #Shape and ShapeDecay have a chromatogram + derivative
        for feature in experiment['features']:
            if feature['type'] in ('Shape', 'ShapeDecay', 'ShapeFront', 'ShapeBack'):
                numPlotsSeq.append(2)
            elif feature['type'] in ('AbsoluteTime', 'AbsoluteHeight'):
                pass
            else:
                numPlotsSeq.append(1)

        numPlots = sum(numPlotsSeq)

        exp_time = target[experimentName]['time']
        exp_value = target[experimentName]['valueFactor']

        fig = figure.Figure(figsize=[10, numPlots*10])
        canvas = FigureCanvas(fig)

        graph_simulations(simulations[experimentName], labels, fig.add_subplot(numPlots, 1, 1))

        graphIdx = 2
        for idx, feature in enumerate(experiment['features']):
            featureName = feature['name']
            featureType = feature['type']

            feat = target[experimentName][featureName]

            selected = feat['selected']
            exp_time = feat['time'][selected]
            exp_value = feat['value'][selected]

            if featureType in ('similarity', 'similarityDecay', 'similarityHybrid', 'similarityHybrid2', 'similarityHybrid2_spline', 'similarityHybridDecay', 
                               'similarityHybridDecay2', 'curve', 'breakthrough', 'dextran', 'dextranHybrid', 'dextranHybrid2', 'dextranHybrid2_spline',
                               'similarityCross', 'similarityCrossDecay', 'breakthroughCross', 'SSE', 'LogSSE', 'breakthroughHybrid', 'breakthroughHybrid2',
                               'Shape', 'ShapeDecay', 'Dextran', 'DextranAngle', 'DextranTest', 'ShapeDecaySimple',
                               'ShapeFront', 'ShapeBack'):
                
                graph = fig.add_subplot(numPlots, 1, graphIdx) #additional +1 added due to the overview plot

                for idx, (sim, label) in enumerate(zip(simulations[experimentName],labels)):
                    sim_time, sim_value = util.get_times_values(sim, target[experimentName][featureName])

                    if idx == 0:
                        linewidth = 2
                    else:
                        linewidth = 1
                    
                    graph.plot(sim_time, sim_value, '--', label=label, color=get_color(idx, len(simulations[experimentName]) + 1, cm_plot), linewidth = linewidth)

                graph.plot(exp_time, exp_value, '-', label='Experiment', color=get_color(len(simulations[experimentName]), len(simulations[experimentName]) + 1, cm_plot), linewidth=2)
                graphIdx += 1
            
            if featureType in ('similarity', 'similarityDecay', 'similarityHybrid', 'similarityHybrid2', 'similarityHybrid2_spline', 'similarityHybridDecay', 
                               'similarityHybridDecay2', 'curve', 'breakthrough', 'dextran', 'dextranHybrid', 'dextranHybrid2', 'dextranHybrid2_spline',
                               'similarityCross', 'similarityCrossDecay', 'breakthroughCross', 'SSE', 'LogSSE', 'breakthroughHybrid', 'breakthroughHybrid2',
                               'Shape', 'ShapeDecay', 'Dextran', 'DextranAngle', 'DextranTest', 'DextranQuad',
                               'Dextran3', 'DextranShape',
                               'ShapeFront', 'ShapeBack'):

                graph = fig.add_subplot(numPlots, 1, graphIdx) #additional +1 added due to the overview plot
                for idx, (sim, label) in enumerate(zip(simulations[experimentName],labels)):
                    sim_time, sim_value = util.get_times_values(sim, target[experimentName][featureName])
                    sim_spline = smoothing.smooth_data_derivative(sim_time, sim_value, feat['critical_frequency'], feat['smoothing_factor'])
                    
                    if idx == 0:
                        linewidth = 2
                    else:
                        linewidth = 1
                    
                    graph.plot(sim_time, sim_spline, '--', label=label, color=get_color(idx, len(simulations[experimentName]) + 1, cm_plot), linewidth = linewidth)

                
                exp_spline = smoothing.smooth_data_derivative(exp_time, exp_value, feat['critical_frequency'], feat['smoothing_factor'])
                graph.plot(exp_time, exp_spline, '-', label='Experiment', color=get_color(len(simulations[experimentName]), len(simulations[experimentName]) + 1, cm_plot), linewidth=2)
                graphIdx += 1
                        
            graph.legend()

        fig.savefig(str(dst)) 

def graph_simulations(simulations, simulation_labels, graph):
    linestyles = ['-', '--', '-.', ':']
    for idx_sim, (simulation, label_sim) in enumerate(zip(simulations, simulation_labels)):

        comps = []

        ncomp = int(simulation.root.input.model.unit_001.ncomp)
        isotherm = bytes(simulation.root.input.model.unit_001.adsorption_model)

        hasSalt = isotherm in saltIsotherms

        solution_times = simulation.root.output.solution.solution_times

        hasColumn = isinstance(simulation.root.output.solution.unit_001.solution_outlet_comp_000, Dict)

        if hasColumn:
            for i in range(ncomp):
                comps.append(simulation.root.output.solution.unit_001['solution_column_outlet_comp_%03d' % i])
        else:
            for i in range(ncomp):
                comps.append(simulation.root.output.solution.unit_001['solution_outlet_comp_%03d' % i])

        if hasSalt:
            graph.set_title("Output")
            graph.plot(solution_times, comps[0], 'b-', label="Salt")
            graph.set_xlabel('time (s)')
        
            # Make the y-axis label, ticks and tick labels match the line color.
            graph.set_ylabel('mM Salt', color='b')
            graph.tick_params('y', colors='b')

            axis2 = graph.twinx()
            for idx, comp in enumerate(comps[1:]):
                axis2.plot(solution_times, comp, linestyles[idx], color=get_color(idx_sim, len(simulation_labels), cm_plot), label="P%s %s" % (idx, label_sim))
            axis2.set_ylabel('mM Protein', color='r')
            axis2.tick_params('y', colors='r')


            lines, labels = graph.get_legend_handles_labels()
            lines2, labels2 = axis2.get_legend_handles_labels()
            axis2.legend(lines + lines2, labels + labels2, loc=0)
        else:
            graph.set_title("Output")
        
            for idx, comp in enumerate(comps):
                graph.plot(solution_times, comp, linestyles[idx], color=get_color(idx_sim, len(simulation_labels), cm_plot), label="P%s %s" % (idx, label_sim))
            graph.set_ylabel('mM Protein', color='r')
            graph.tick_params('y', colors='r')
            graph.set_xlabel('time (s)')

            lines, labels = graph.get_legend_handles_labels()
            graph.legend(lines, labels, loc=0)

def main():
    cache.setup_dir(sys.argv[1])
    util.setupLog(cache.settings['resultsDirLog'], "mle.log")
    cache.setup(sys.argv[1])

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    with h5py.File(mcmc_h5, 'r') as h5:
        flat_chain = h5['/flat_chain'][()]
        mcmc_acceptance = h5['/mcmc_acceptance'][()]
        process_mle(flat_chain, len(mcmc_acceptance), cache)

if __name__ == "__main__":
    main()



import CADETMatch.util as util
import numpy

name = "keq"
count = 2
count_extended = 3

def getUnit(location):
    return location[0].split('/')[3]

def transform(parameter):
    def trans_a(i):
        return numpy.log(i)

    def trans_b(i):
        return numpy.log(i)

    return [trans_a, trans_b]

def untransform(seq, cache, parameter):
    values = [numpy.exp(seq[0]), numpy.exp(seq[0])/(numpy.exp(seq[1]))]
    headerValues = [values[0], values[1], values[0]/values[1]]
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    values = untransform_matrix_inputorder(matrix, cache, parameter)
    values[:,1] = values[:,0] / values[:,1]
    return values

def untransform_matrix_inputorder(matrix, cache, parameter):
    values = numpy.exp(matrix)
    return values

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)

    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        location = parameter['location']
    
        comp = parameter['component']
        bound = parameter['bound']
    
        unit = getUnit(location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        position = boundOffset[comp] + bound
        sim[location[0].lower()][position] = values[0]
        sim[location[1].lower()][position] = values[1]

    return values, headerValues

def setupTarget(parameter):
    location = parameter['location']
    bound = parameter['bound']
    comp = parameter['component']

    sensitivityOk = 1
    nameKA = location[0].rsplit('/', 1)[-1]
    nameKD = location[1].rsplit('/', 1)[-1]
    unit = int(location[0].split('/')[3].replace('unit_', ''))

    return [(nameKA, unit, comp, bound), (nameKD, unit, comp, bound)], sensitivityOk

def getBounds(parameter):
    minKA = parameter['minKA']
    maxKA = parameter['maxKA']
    minKEQ = parameter['minKEQ']
    maxKEQ = parameter['maxKEQ']

    minValues = numpy.log([minKA, minKEQ])
    maxValues = numpy.log([maxKA, maxKEQ])

    return minValues, maxValues

def getHeaders(parameter):
    location = parameter['location']
    nameKA = location[0].rsplit('/', 1)[-1]
    nameKD = location[1].rsplit('/', 1)[-1]
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
    headers.append("%s Comp:%s Bound:%s" % (nameKD, comp, bound))
    headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
    return headers

def getHeadersActual(parameter):
    location = parameter['location']
    nameKA = location[0].rsplit('/', 1)[-1]
    nameKD = location[1].rsplit('/', 1)[-1]
    bound = parameter['bound']
    comp = parameter['component']
    
    headers = []
    headers.append("%s Comp:%s Bound:%s" % (nameKA, comp, bound))
    headers.append("%s/%s Comp:%s Bound:%s" % (nameKA, nameKD, comp, bound))
    return headers

def setBounds(parameter, lb, ub):
    parameter['minKA'] = lb[0]
    parameter['maxKA'] = ub[0]
    parameter['minKEQ'] = lb[2]
    parameter['maxKEQ'] = ub[2]

